import json

def init_configuration(filename: str, outDir: str):
  """
    Create kamino initial configurations
  """
  repo_name = filename
  initial_configs = {
    'project': {
      'name': repo_name
    },
    'ignore': ['*node_modules*', '*__pycache__*', '.git*', '.pytest_cache*'],
    'tags': {
      'exampleTag': 'value' 
    }
  }

  json_initial_configs = json.dumps(initial_configs, indent=2)
  with open(f'{outDir}/{filename}.kamino.json', 'w', encoding="utf-8") as file:
    file.write(json_initial_configs)
    print("Configuration file created!")


import argparse
import os
import sys

from kamino.cli import (
  init,
  run
)


def create_argument_parser() -> argparse.ArgumentParser:
  parser = argparse.ArgumentParser(
    prog="kamino",
    description="An easy repository replication tool"
  )

  subparsers = parser.add_subparsers(help="kamino commands")
  init.add_subparser(subparsers)
  run.add_subparser(subparsers)

  return parser

def main():
  arg_parser = create_argument_parser()
  cmdline_arguments = arg_parser.parse_args()

  # insert current path in syspath so custom modules are found
  sys.path.insert(1, os.getcwd())
  try:
    if hasattr(cmdline_arguments, "func"):
        cmdline_arguments.func(cmdline_arguments)
  except Exception as e:
    print(e)

if __name__ == "__main__":
  main()


import re
from typing import Dict
def replace_tags(content: str, patterns: Dict) -> str:
  result = content
  for pattern, pattern_value in patterns.items():
    regex = fr"\${pattern}\$"
    regex = re.compile(regex)
    result = re.sub(regex, pattern_value, result)
  return result



from kamino import load_repository, save_repository
from kamino.constants import DEFAULT_CONFIG_FILE, DEFAULT_CONFIG_NAME
from ..core.traverse_directory import ReplaceAllTags
from ..core.load_configuration import load_configuration
from . import *

def set_run_arguments(parser: argparse.ArgumentParser):
  parser.add_argument(
    "-d",
    "--directory",
    type=str,
    default='.',
    help="Directory of where kamino must execute the replication."
  )

  parser.add_argument(
    "-c",
    "--config",
    type=str,
    default=DEFAULT_CONFIG_FILE,
    help="Path to the configuration file that kamino should use to replicate."
  )
  parser.add_argument(
    "-o",
    "--out",
    type=str,
    default='./kamino',
    help="Directory of where kamino should send the replication result."
  )

def add_subparser(subparsers: SubParsersAction):
  run_parser = subparsers.add_parser(
  "run",
  formatter_class=argparse.ArgumentDefaultsHelpFormatter,
  help="Executes kamino's replication."
  )
  run_parser.set_defaults(func=run)

  set_run_arguments(run_parser)

def run(args: argparse.Namespace):
  """
  1. verifica se há um arquivo de config
    - caso não haja arquivo especificado, usar padrão
    - caso haja arquivo especificado, usar o mesmo
  2. validar arquivo de config
  3. carrega o arquivo de config com base no argumento
  4. carrega repositório
    - se configuracao possuir ignore patterns, não carregar diretórios/arquivos que se encaixam nos patterns
  5. executa troca de tags por seus valores
  6. salva novo repositório no caminho especificado
    - se não houver caminho especificado, usar saída padrão './out'
    - se houver caminho de saída especificado, usar saída especificada
  """
  print("loading configurations...")
  configurations = load_configuration(args.config)
  print("configurations loaded.")

  print("loading repository...")
  repository = load_repository(args.directory, configurations['ignore'])
  print("repository loaded.")

  replacer = ReplaceAllTags(configurations['tags'])
  replacer.traverse_repository(repository)
  print("Saving repository")
  save_repository(repository, args.out, configurations['project']['name'])
  print("Repository Replication concluded!!!")


import os
from abc import ABC
import locale
from typing import Dict

from kamino.constants import DEFAULT_OUT_DIR

from .replace_tags import replace_tags

from .respository.Directory import Directory
from .respository.File import File
from .respository.Repository import Repository


class TraverseDirectory(ABC):

  def traverse_repository(self, repository: Directory):

    for repository_item in repository._children:
      if type(repository_item) == Directory:
        self.before_enter_recursion(repository_item)
        self.traverse_repository(repository_item)
        self.after_exit_recursion(repository_item)
      self.finalize_traverse_hook(repository_item)

  def before_enter_recursion(self, repository_item):
    ...

  def after_exit_recursion(self, repository_item):
    ...
  
  def finalize_traverse_hook(self, repository_item):
    ...


class CreateRepository(TraverseDirectory):

  def create_repository(self, repository_item: Directory, out: str, repo_name: str = None):
    out_dir = Directory(out)
    if repo_name is not None:
      repository_item.name = repo_name

    out_dir.add(repository_item)

    self.create_dir(out_dir)
    self.traverse_repository(out_dir)

  def before_enter_recursion(self, repository_item: Directory):
    self.create_dir(repository_item)

  def finalize_traverse_hook(self, repository_item: Repository):
    if type(repository_item) == File:
      self.create_file(repository_item)

  def create_file(self, file: File):
    mode = 'wb' if type(file.content) is bytes else 'w'
    encoding= locale.getpreferredencoding() if type(file.content) is bytes else None
    with open(file.get_path(), mode, encoding=encoding) as new_file:
      new_file.write(file.content)

  def create_dir(self, dir: Directory): 
    if not os.path.exists(dir.get_path()):
      os.makedirs(dir.get_path())


class ReplaceAllTags(TraverseDirectory):

  def __init__(self, patterns: Dict) -> None:
    self.patterns = patterns
    self.ignore_extensions = [".pdf", ".jpg", ".jpeg", ".svg", ".ico", ".png", ".jar", ".keystore", ".bat", ".exe", '.ttf']
    super().__init__()

  def finalize_traverse_hook(self, repository_item: Repository):
    if type(repository_item) == File:
      if repository_item.get_extension() not in self.ignore_extensions:
        print(f"Replacing tags in {repository_item.get_path()}")
        repository_item.content = replace_tags(repository_item.content, self.patterns)


from .core.init_configuration import init_configuration
from .core.load_configuration import load_configuration
from .core.load_repository import load_repository
from .core.save_repository import save_repository


from .respository import Directory
from .traverse_directory import CreateRepository


def save_repository(repository: Directory, path: str = '.', repo_name: str = None):
  """
    Salva a arvore de diretórios no caminho especificado.
  """
  creator = CreateRepository()
  creator.create_repository(repository, path, repo_name)




from __future__ import annotations
from abc import ABC
from dataclasses import dataclass
from typing import List


@dataclass
class Repository(ABC):
  name: str
  parent: Repository = None
  content: str = None

  def get_path(self) -> str:
    return ''

  def get_dirs_list(self) -> List[str]:
    return []

  def is_composite(self) -> bool:
        return False

  def add(self, component: Repository) -> None:
      pass

  def remove(self, component: Repository) -> None:
      pass




from fnmatch import fnmatch
import os
from typing import List

from kamino.constants import DEFAULT_RUN_DIR

from .respository.File import File
from .respository.Directory import Directory

def climb(dir_tree: Directory) ->Directory:
  """
    Sobe um diretório na hierarquia de pastas
  """
  if dir_tree.parent is not None:
    return dir_tree.parent
  return dir_tree

def climb_tree(dir_tree: Directory, new_dir_name: str, path: str) ->Directory:
  
  while f"{dir_tree.get_path()}{os.sep}{new_dir_name}" != path:
    dir_tree = climb(dir_tree)
  return dir_tree

def create_file(dirpath: str, filename: str):
  fname = os.path.join(dirpath,filename)

  try:
    with open(fname, encoding="utf-8") as myfile:
      result = File(name=filename, content=myfile.read())
  except:
    with open(fname, 'rb') as myfile:
      result = File(name=filename, content=myfile.read())
  return result

def create_dir(parent_dir: Directory, dir_name: str) -> Directory:
  """
    Cria um diretório e delega um diretório pai a ele.
  """
  new_dir = Directory(name=dir_name)
  parent_dir.add(new_dir)
  return new_dir

def match_patterns(patterns: List[str], to_match: str) -> bool:
  """
    checks if a file matches a unix pattern
  """
  result = False
  for pattern in patterns:
    result = fnmatch(to_match, pattern)
    if result ==True: 
      break
  
  return result

def remove_from_loading(patterns: List[str], items: List[str], dirpath: str):
  for item in items:
    itempath=os.path.join(dirpath, item)
    if match_patterns(patterns, itempath):
      items.remove(item)
  return items


def load_repository(root_dir: str, patterns: List[str]) -> Directory:
  if root_dir == DEFAULT_RUN_DIR:
    root_dir_name = os.path.basename(os.getcwd())
  else:
    root_dir_name = root_dir.split(os.sep)[-1]

  root = Directory(name='.')
  tree_head = root

  for dirpath, dirs, files in os.walk(f"{root_dir}"):	
    dirs_list = dirpath.split(os.sep)
    current_dir = dirs_list[-1]

    if dirpath != f"{root_dir}" and not match_patterns(patterns, dirpath):
      tree_head = climb_tree(tree_head, current_dir, dirpath)
      print(f"Copying dir: {dirpath}")
      tree_head = create_dir(tree_head, current_dir)

    for filename in files:
      if not match_patterns(patterns, os.path.join(dirpath,filename)):
        print(f"Reading: {filename}")
        new_file = create_file(dirpath, filename)
        tree_head.add(new_file)

  root.name = root_dir_name
  return root



import json
import locale
from jsonschema import validate
from typing import Dict

configuration_schema = {
  "type": "object",
  "properties": {
    "project": {
      "type": "object",
      "properties": {
        "name": {
          "type": "string"
        },
        "outDir": {
          "type": "string"
          }
      }
    },
    "ignore": {"type": "array"},
    "tags": {"type": "object"}
  },
  "required": ["ignore", "tags"]
}

def check_configurations(configuration: Dict):
  validate(configuration, configuration_schema)

def load_configuration(filePath: str)-> Dict:
  configuration = {}
  with open(filePath, encoding=locale.getpreferredencoding()) as configurationFile:
    configuration = json.load(configurationFile)
  check_configurations(configuration)

  configuration['ignore'].append("*kamino*")
  return configuration


from __future__ import annotations

import os
from typing import List
from .Repository import Repository
from dataclasses import dataclass, field

@dataclass
class Directory(Repository):
    """
    Composite class que representa um diretório. O diretório pode conter vários filhos
    sendo eles outros diretórios ou arquivos.
    """
    _children: List[Repository] = field(default_factory=lambda : [])


    def add(self, component: Repository) -> None:
        """
        Um objeto composite pode adicionar ou remover outros componentes.
        """
        self._children.append(component)
        component.parent = self

    def remove(self, component: Repository) -> None:
        self._children.remove(component)
        component.parent = None

    def is_composite(self) -> bool:
        return True

    def get_path(self) -> str:
        if self.parent != None:
            return f"{self.parent.get_path()}{os.sep}{self.name}"
        return f"{self.name}"

    def get_dirs_list(self) -> List[str]:
        if self.parent != None:
            result = self.parent.get_dirs_list()
            result.append(self.name)
            return result 
        return [f"{self.name}"]



import os
from typing import List
from .Repository import Repository
class File(Repository):
    """
    The Leaf class represents the end objects of a composition. A leaf can't
    have any children.

    Usually, it's the Leaf objects that do the actual work, whereas Composite
    objects only delegate to their sub-components.
    """

    def get_extension(self):
        name_as_list = self.name.split('.')
        return f".{name_as_list[-1]}"

    def get_path(self) -> str :
        if self.parent == None:
            return f"./{self.name}"
        return f"{self.parent.get_path()}{os.sep}{self.name}"

    def get_dirs_list(self) -> List[str]:
        result = self.parent.get_dirs_list()
        result.append(self.name)
        return result



from argparse import _SubParsersAction
import argparse

from kamino.constants import DEFAULT_CONFIG_NAME, DEFAULT_OUT_DIR

def set_init_arguments(parser: argparse.ArgumentParser):
  parser.add_argument("-f", "--filename", type=str, default=DEFAULT_CONFIG_NAME, help="Name of the config file.")
  parser.add_argument("-o", "--outDir", type=str, default=DEFAULT_OUT_DIR, help="Exit directory.")

def add_subparser(
    subparsers: _SubParsersAction
) -> None:
  """
  add all init parsers.

    Args:
      subparsers: subparser we are going to attach to
  """
  init_parser = subparsers.add_parser(
    "init",
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    help="Creates an initial json configuration containing 'ignore' and 'tags' keys"
    )
  init_parser.set_defaults(func=init)

  set_init_arguments(init_parser)




def init(args: argparse.Namespace) -> None:
  """
  Creates a new configuration json file

  Args:
    filename: name of the configuration file
    outDir: name of the directory to save the configurations
  """
  from kamino import init_configuration
  init_configuration(args.filename, args.outDir)

